# mautrix-instagram - A Matrix-Instagram puppeting bridge.
# Copyright (C) 2020 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from ..types import UserSearchResponse
from .base import BaseAndroidAPI


class UserAPI(BaseAndroidAPI):
    async def search_users(self, query: str, count: int = 30) -> UserSearchResponse:
        req = {
            "timezone_offset": self.state.device.timezone_offset,
            "q": query,
            "count": count,
        }
        return await self.std_http_get(
            "/api/v1/users/search/", query=req, response_type=UserSearchResponse
        )
